//*****************************************************************************
//
// Author		: Konstantin Chizhov
// Date			: 2010
// All rights reserved.

// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.

// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation and/or
// other materials provided with the distribution.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
// OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//*****************************************************************************

//*****************************************************************************
// 7segments.h
//*****************************************************************************
//                                            Vcc             Vcc
//                                             |               |
//                                         __|/            __|/
//        |=============================  |  |\. =========|  |\.   ...
//        |       |=======================|    |               |
//        |       |                            |
//       ___________                                       _GFE'DCBA notation
//      |           |                ******************************************
//      |           |==[330]==[G]==<==      CA |           _011 1111  0x3F [0]
//      |           |                                      _000 0110  0x06 [1]
//      |           |==[330]==[F]==<==     ____A____       _101 1011  0x5B [2]
//      |           |                     |         |      _100 1111  0x4F [3]
//      |           |==[330]==[E]==<==   F|         |B     _110 0110  0x66 [4]
//      |           |                     |         |      _110 1101  0x6D [5]
//      |           |==[330]==[D]==<==    |____G___ |      _111 1101  0x7D [6]
//      |           |                     |         |      _000 0111  0x07 [7]
//      |     uC    |==[330]==[C]==<==   E|         |C     _111 1111   x7F [8]
//      |           |                     |         |      _110 1111  0x6F [9]
//      |           |==[330]==[B]==<==    |____D____|/o/   _111 0111  0x77 [A]
//      |           |                                      _111 1100  0x7C [b]
//      |           |==[330]==[A]==<==    7 Segment CA     _011 1001  0x39 [C]
//      |           |                                      _101 1110  0x5E [d]
//      |           |==[330]==[o]==<==                     _111 1001  0x79 [E]
//      |           |                                      _111 0001  0x71 [F]
//      |___________|                *******************************************
//
//******************************************************************************
//       ___________                                       _GFE'DCBA notation
//      |           |                ******************************************
//      |           |==[330]==[G]==>==    7 Segment CC     _011 1111  0x3F [0]
//      |           |                                      _000 0110  0x06 [1]
//      |           |==[330]==[F]==>==     ____A____       _101 1011  0x5B [2]
//      |           |                     |         |      _100 1111  0x4F [3]
//      |           |==[330]==[E]==>==   F|         |B     _110 0110  0x66 [4]
//      |           |                     |         |      _110 1101  0x6D [5]
//      |           |==[330]==[D]==>==    |____G___ |      _111 1101  0x7D [6]
//      |           |                     |         |      _000 0111  0x07 [7]
//      |     uC    |==[330]==[C]==>==   E|         |C     _111 1111  0x7F [8]
//      |           |                     |         |      _110 1111  0x6F [9]
//      |           |==[330]==[B]==>==    |____D____|/o/   _111 0111  0x77 [A]
//      |           |                                      _111 1100  0x7C [b]
//      |           |==[330]==[A]==>==         |           _011 1001  0x39 [C]
//      |           |                          |           _101 1110  0x5E [d]
//      |           |==[330]==[o]==>==      CC |           _111 1001  0x79 [E]
//      |           |                          |           _111 0001  0x71 [F]
//      |___________|                **********|********************************
//        |       |                        __|/
//        |       |                       |  |\.               |
//        |       |=======================|    |           __|/
//        |                                  GND          |  |\..
//        |===============================================|    |       ...
//                                                            GND
//
//******************************************************************************
// Although each I/O port can sink more than the test conditions (20 mA at
// VCC = 5V, 10 mA at VCC = 3V) under steady state conditions (non-transient),
// the following must be observed:
//
// ATmega48A/PA/88A/PA/168A/PA/328/P:  see p.304 pt.4.
// *********************************
// 1] The sum of all IOL, for ports C0 - C5, ADC7, ADC6 should not exceed 100 mA.
// 2] The sum of all IOL, for ports B0 - B5, D5 - D7, XTAL1, XTAL2 should not 7
//    exceed 100 mA.
// 3] The sum of all IOL, for ports D0 - D4, RESET should not exceed 100 mA.
//
// If IOL exceeds the test condition, VOL may exceed the related specification.
// Pins are not guaranteed to sink current greater than the listed test condition.
//
//*****************************************************************************
	#ifndef SEVEN_SEGMENTS_H
	#define SEVEN_SEGMENTS_H

	#include <stddef.h>

	#include <static_if.h>
	#include <flashptr.h>
//#include <template_utils.h>

	#include <drivers/c7segments_.h>

	namespace Mcucpp {
//*****************************************************************************

	class cCharMap7Segment {                              // _GFE'DCBA notation
//*****************************************************************************
// obsolete, replaced by c7Segments_ :              add_char | charMap(char)
//*****************************************************************************
	enum {test=1, test_=2};
	
	public:
		static uint8_t charMap (uint8_t bcd) {
			static const FLASH_STORAGE uint8_t table[] =   // !!! HEX-STRING !!!
					"\x3f\x06\x5b\x4f\x66\x6d\x7d\x7\x7f\x6f\x77\x7c\x39\x5e\x79\x71";

			FLASH_PTR(uint8_t) tablePtr(table);
			return  *(tablePtr + size_t(bcd));
		}

		static uint8_t Minus(){return 0x40;}   // G
		static uint8_t Empty(){return 0x00;}   // none
		static uint8_t Exp()  {return 0x23;}   // FBA
	};


	class cBase7Segment {
//*************************************************************************
//	cBase7Segment();           // constructor

	public:
		enum FeaturesT {
			Seg7None     = 0,        // ??
			Seg7BigNum   = 1,        // ??
			Seg7Align    = 2,        // ??
			Seg7All      = 3         // all segments ?
		};

		enum FlagsT    {
			DecFlag      = 0,        // decimal     numbers
			HexFlag      = 1,        // hexadecimal numbers
			LeftAligned  = 2,        // left        aligned
			RightAligned = 0         // right       aligned
		};
	};

template < class Segments,                              // _GFE'DCBA  // PinList Segments
					 class Columns,                               //  digits    // PinList Columns
					 cBase7Segment::FeaturesT Features = cBase7Segment::Seg7All,// feature
					 class CharMap = c7Segments_                                // new CharTable
//				 class CharMap = cCharMap7Segment                           // old CharTable
//**********************************************************************
>	class c7Segment: public cBase7Segment {
//**********************************************************************
//	STATIC_ASSERT(Segments::Length == 7 || Segments::Length == 8);
//	STATIC_ASSERT( Columns::Length >= 2);
	
		typedef typename StaticIf <
								Columns::Length <= 2, uint8_t,
										typename StaticIf < 
												Columns::Length <= 4, uint16_t,
														typename StaticIf <
																Columns::Length <= 8, uint32_t, uint64_t
														>::Result 
										>::Result
						>::Result MaxDisplayNumberT;  // Result: uint8_t..uint64_t
	
	public:
		enum{ Digits                   = Columns::Length};
		enum{ MaxDisplayNumber         = Util::Pow<10, Digits  >::value-1};
		enum{ MaxDisplayNumberWithSign = Util::Pow<10, Digits-1>::value-1};
		enum{ MaxDisplayNumberHex      = Util::Pow<16, Digits  >::value-1};

		static void Update();
		static void LeftAlign() {sdata.flags |=  LeftAligned ;}
		static void RightAlign(){sdata.flags &= ~LeftAligned ;}
		static void Hex()       {sdata.flags |=  HexFlag     ;}
		static void Dec()       {sdata.flags &= ~DecFlag     ;}

		template<class T>
		static void Write(T value) {
	//************************************************************
	// write data into sPrivateData
	//************************************************************
			bool minusSign = false;
			if(Util::IsSigned<T>::value && !(sdata.flags & HexFlag) ) {
				if(value < 0) {
					value = -value;
					minusSign = true;
				}
			}

			typedef typename Util::Unsigned<T>::Result UT;
			
			UT       uvalue =  static_cast<UT>(value);
			uint8_t  exp    =  0;
			uint8_t  base   =  sdata.flags & HexFlag ? 16  // true
																							 : 10; // false
			uint8_t *ptr    = &sdata.value[0] + Digits;

			MaxDisplayNumberT         maxNum = MaxDisplayNumber;
			if(minusSign)             maxNum = MaxDisplayNumberWithSign;
			if(sdata.flags & HexFlag) maxNum = MaxDisplayNumberHex;

			do {                                          // ??
		//************************************************
				T q = uvalue / base;
				if(uvalue >= maxNum) exp++;
				else{ uint8_t rem = uint8_t(uvalue - q * base);
					         *--ptr = CharMap::charMap(rem);
				}
				uvalue = q;
		//************************************************
			} while (uvalue != 0);

			if(minusSign) *--ptr = CharMap::charMap(minus);

			if((Features & cBase7Segment::Seg7BigNum) && Digits > 2 && exp) {
		//**************************************************************
		//
		//**************************************************************
				ptr = &sdata.value[0] + Digits;
				if(exp >= base - 1) exp++;
				exp++;

				do {                                   // ??
			//*****************************************
					uint8_t  eq = exp / base;
					uint8_t rem = uint8_t(exp - eq*base);
					*--ptr = CharMap::charMap(rem);
					exp = eq;
			//*****************************************
				} while (exp != 0);
				
				*--ptr = CharMap::charMap(exp);
			}
			else { // ((Features & Base7Segment::Seg7BigNum) && Digits > 2 && exp)
		//**********************************************************************
		//
		//**********************************************************************
				uint8_t shift = uint8_t(ptr - &sdata.value[0]);
				uint8_t outputSize = Digits - shift;
				
				if((Features & cBase7Segment::Seg7Align) && sdata.flags & LeftAligned) {
			//*********************************************************************
			//
			//*********************************************************************
					for(uint8_t i=0; i < outputSize; i++) 
						sdata.value[i] = sdata.value[i + shift];
					ptr = &sdata.value[0] + Digits;
				}
				for(uint8_t i=0; i < shift; i++) 
					*--ptr = CharMap::charMap(space);
			}
		}
	
	private:
		struct sPrivateData {    
	//*****************************
			uint8_t flags;           //
			uint8_t position;        //
			uint8_t value[Digits];   //
		};
		
		static sPrivateData sdata; // data is stored here 
	//*****************************
	};


// init sPrivateData
//*****************************************************************************
template < class Segments, 
					 class Columns, 
					 cBase7Segment::FeaturesT Features,
					 class CharMap
//*****************************************
>	class c7Segment<Segments, Columns, Features, CharMap>::sPrivateData
					c7Segment<Segments, Columns, Features, CharMap>::sdata = {
	//*******************************************************#####**
				c7Segment<Segments, Columns, Features, CharMap>::DecFlag,0,{0}
	};

// update display
//*****************************************************************************
template < class Segments,
					 class Columns, 
					 cBase7Segment::FeaturesT Features,
					 class CharMap
//*****************************************
>	void c7Segment<Segments, Columns, Features, CharMap>::Update() {
//******************************************************######**
//  uses : pinlist.h
// *************************************************************
//  template<class PINS>
//  void PinSet<PINS>::Write(typename PinSet<PINS>::uIntType value) {
//  IoPrivate::PortWriteIterator<Ports, PINS, uIntType>::Write(value);
//  }
//**************************************************************
		Columns ::template SetConfiguration<Columns::Out>();                         // dir
		Columns::          Write  (typename Columns::uIntType(1) << sdata.position); // pos
	
		Segments::template SetConfiguration<Segments::Out>();                        // dir 
		Segments::         Write(sdata.value[sdata.position++]);                     // data
		if(sdata.position >= Digits)         sdata.position=0  ;                     // CR
	}
	//_________________________________________________________________________
} //.Mcucpp
#endif
