//*****************************************************************************
// cMax7219                      standard place to store : ..\mcucpp\drivers
//*****************************************************************************
// based on max7219.c        (c) Peter Dannegger
//*****************************************************************************

//                                MAX7219/21
//                               ************
//       ___________                                      |-ABC'DEFG| notation
//      |           |                ******************************************
//      |           |==[330]==[G]==>==    7 Segment CC    |_111 1110| 0x7E [0]
//      |           |                                     |_011 0000| 0x30 [1]
//      |           |==[330]==[F]==>==     ____A____      |_110 1101| 0x6D [2]
//      |           |                     |         |     |_111 1001| 0x79 [3]
//      |           |==[330]==[E]==>==   F|         |B    |_011 0011| 0x33 [4]
//      |           |                     |         |     |_101 1011| 0x5B [5]
//      |           |==[330]==[D]==>==    |____G___ |     |_101 1111| 0x5F [6]
//      |           |                     |         |     |_111 0000| 0x70 [7]
//      |           |==[330]==[C]==>==   E|         |C    |_111 1111| 0x7F [8]
//      |  MAX7219  |                     |         |     |_111 1101| 0x7B [9]
//      |           |==[330]==[B]==>==    |____D____|/o/  |_111 0111| 0x77 [A]
//  ==>=|CLK        |                                     |_001 1111| 0x1F [b]
//      |           |==[330]==[A]==>==         |          |_010 0111| 0x4E [C]
//  ==>=|DIN        |                          |          |_011 1101| 0x3D [d]
//      |           |==[330]==[o]==>==      CC |          |_100 1111| 0x4F [E]
//  ==>=|LOAD       |                          |          |_100 0111| 0x47 [F]
//      |___________|                **********|*******************************
//        |       |                            |               |
//        |       |============================|               |
//        |                                                    |
//        |====================================================|  ...
//
//
/******************************************************************************
 >>>        ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___ ___     >>>
 DATA _____|D00|D01|D02|D03|D04|D05|D06|D07|D08|D09|D10|D11|D12|D13|D14|D15|____
 >>>  ______  ^_  ^_  ^_  ^_  ^_  ^_  ^_  ^_  ^_  ^_  ^_  ^_  ^_  ^_  ^_  ^_____
 CLK        |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_| |_|     >>>
 >>> _____^  _______________________________________________________________ ___
 LOAD     |_|                                                               |   >>>
 
******************************************************************************/
	#ifndef MAX7219_HPP
	#define Max7219_HPP 

	#include <delay.h>
	#include <c7segments_.h>             // add_char | CharMap(char)
	
	namespace Mcucpp {
//****************************************************************************


	template < class DIN,                // serial data input [Max7219:PIN 1] TPin
						 class LOAD,               // load data/CS      [        PIN12] TPin
						 class CLK,                // clock             [        PIN13] TPin
						 uint8_t digits    = 8,    //
						 uint8_t intensity = 5,    //
 						 uint8_t decode    = 0     //
//***************************************
>	class cMax7219 {
//***************************************

		typedef IO::PinList < DIN, LOAD, CLK > CtrlBus;
//************************************************

		static uint8_t displayMem[digits];
//***************************************

	public:
		static uint16_t Display  (uint8_t mode) { return 0x0C00 | mode;}; // mode 0|1
		static uint16_t Test     (uint8_t mode) { return 0x0F00 | mode;}; // mode 0|1
		static uint16_t Digits   (uint8_t mode) { return 0x0B00 | mode;}; // mode 0|1|2|3|4|5|6|7
		static uint16_t Decode   (uint8_t mode) { return 0x0900 | mode;}; // mode 0|1|F|FF|
		static uint16_t Intensity(uint8_t mode) { return 0x0A00 | mode;}; // mode 0|1|2|3|4|5|6|7|8|9|A|B|C|D|E|F
//**********************************************************************

	//	cMax7219() { }                         // constructor
	//*******************************************

		static void Init() { 
	//**************************************
			CtrlBus::template
				SetConfiguration<CtrlBus::Out>();
			init();
		}

	__attribute__ ((noinline))
		static void write( uint16_t data ) {     // write serial data 
	//*******************************************
			uint8_t i;

			CtrlBus::template 
				SetConfiguration<CtrlBus::Out>();

			LOAD::Clear();                         // \__
		//*****************************************

			for( i = 12; i; i-- ) {   // MSB first // bit 0..11 // bit 12..15 unused
		//*****************************************
				CLK::Clear();                        // \__
//				delay_us< 5,F_CPU>(  );            // 5us
				DIN::Set();                          //
//				delay_us< 5,F_CPU>(  );            // 5us
				if((data & 0x0800)==0) DIN::Clear(); // mask for |BIT.11|
				data<<=1;                            // shift next input bit
				CLK::Set();                          // shift data into max7219
//				delay_us< 5,F_CPU>(  );            // 5us
			}
			LOAD::Set();                           // latch data in max7219
		}
		
		static void init( void ) { 
	//*******************************************
			write   ( Display   (1));              // display on
			write   ( Test      (1));              // testmode on
			delay_ms< 200,F_CPU>(  );              // 200ms 
			write   ( Test      (0));              // testmode off
			write   ( Decode    (decode));         // decode
			write   ( Digits    (digits-1));       // digits
			write   ( Intensity (intensity));      // brightness
//		write   ( Display   (0));              // display off
		} 

		static uint8_t GetDisplayMem (uint8_t i) {
	//****************************************
			return displayMem[i];
		}

		static void SetDisplayMem ( uint8_t i, uint8_t val ) {
	//***************************************************
			displayMem[i] = val;
		}
	
		static void Update( void ) {             // update display data
	//*******************************************
			uint8_t i, j;
	
			write( Display(1));

			for( i = digits; i; i-- ){
		//****************************************************
				j =  c7Segments_::charMap(GetDisplayMem(i-1)&0x3f);
				if (GetDisplayMem(i-1)&0x80) j |= decPoint;
				write(i*256 + j);                    // digit + data
			}
	}

		static void clear( void ) {
	//*****************************************
			uint8_t i;
			for(i=digits;i;i--) SetDisplayMem(i,0);
		}

		static void off( void )  {
	//*****************************************
			write( Display(0));
		}
	};
//____________________________________________________________________________
} //.Mcucpp
#endif
