//****************************************************************************
// cTimebaseDCF77
//****************************************************************************
// based on timebase.c       (c) Peter Dannegger
//****************************************************************************
	#ifndef TIMEBASE_HPP
	#define TIMEBASE_HPP
	
	#include <timer0.h>
	
	namespace Mcucpp { namespace Timers {
//****************************************************************************

	class cTimeBase : public Timer0 {
//****************************************************************************
		enum {                                                        // F_CPU = 14745600 Hz
	//****************************************************************
			T0_COUNT   = (F_CPU / 1024  / 64                         ), // 225[.00000000]divT/sec [reload PER TICK  ]
			T0_SEC_ERR = (F_CPU - 1024L * 64 * T0_COUNT              ), //   0  zu wenig   qT/sec
			T0_SECOND  = (T0_COUNT + T0_SEC_ERR / 1024               ), // 225                    [reload PER SECOND]
			T0_MIN_ERR = (F_CPU - 1024 * (63 * T0_COUNT + T0_SECOND) ), //   0
			T0_MINUTE  = (T0_SECOND + (T0_MIN_ERR * 60 + 512) / 1024 )  // 225 @ 14745600 Hz      [reload PER MINUTE]
		};                                                            // xxx @ 14745xxx Hz      [     xx sec/Monat]

	public:
		static uint8_t _t64count;                    //  64 reloads = 1sec
		static uint8_t _timeflags;                   //     ..|MIN|SEC|TT|

		enum { ONE_TTICK  = 0,
					 ONE_SECOND = 1,
					 ONE_MINUTE = 2
		};    

	//cTimeBase() { }                              // constructor// UNUSED
	//***********************************************

		static void Init() {
	//***********************************************
			Start(Div1024);                            // Clock DIV1024
			EnableInterrupt();                         // Timer0 OverFlow Int Enable
	//***********************************************
	//	_t64count =0;                              // extern init. mandatory
	//	_timeflags=0;                              // extern init. mandatory
		}
		
		static void sync_sec( ) {                    // synchronize by DCF77
	//***********************************************
			Set((uint16_t)(256-T0_MINUTE));            // set Timer0 

			_t64count  = 0;                            // T64Counter - 64 reloads = one sec
			_timeflags = 0;                            // [....'.MST] - MIN,SEC,TICK
		}

		static void Update() {                             
	//***********************************************
	// reload Timer0                               // time base 1 second
	//***********************************************
	// resets the counter for a PRECISE second
	//***********************************************
	// if t64count & 0x3f |xx11'1111| -> TRUE
	//***********************************************
	//     t64count = 64  |0100'0000| -> FALSE > 1s
	//             = 128  |1000'0000| -> FALSE > 1s
	//             = 192  |1100'0000| -> FALSE > 1s
	//             = 256 1|0000'0000| -> FALSE > 1s
	//***********************************************
			TCNT0 = (uint16_t)(256 - T0_COUNT);        // reload PER TICK
			if( ++_t64count & 0x3F ){                  // 64 reloads = one second
				_timeflags = 1<<ONE_TTICK;               // one tick over
				return;
			}

			TCNT0 = (uint16_t )(256 - T0_SECOND);      // reload PER SECOND
			if( _timeflags & (1<< ONE_MINUTE ))
				TCNT0 = (uint16_t )(256 - T0_MINUTE);    // reload PER MINUTE
			_timeflags = 1<<ONE_SECOND^1<<ONE_TTICK;   // one tick, one second over
		}
	};
//____________________________________________________________________________
	}} //.Timers//.Mcucpp
	
	#endif //._TIMEBASEDCF77_
